#include "spirv_internal.hpp"

_SPIRV_OP_INTERNAL(Forward, internal::OpForward)
_SPIRV_OP_INTERNAL(TypeTokenINTEL, internal::OpTypeTokenINTEL)
_SPIRV_OP_INTERNAL(JointMatrixGetElementCoordINTEL,
                   internal::OpJointMatrixGetElementCoordINTEL)
_SPIRV_OP_INTERNAL(CooperativeMatrixPrefetchINTEL,
                   internal::OpCooperativeMatrixPrefetchINTEL)
_SPIRV_OP_INTERNAL(CooperativeMatrixLoadCheckedINTEL,
                   internal::OpCooperativeMatrixLoadCheckedINTEL)
_SPIRV_OP_INTERNAL(CooperativeMatrixStoreCheckedINTEL,
                   internal::OpCooperativeMatrixStoreCheckedINTEL)
_SPIRV_OP_INTERNAL(CooperativeMatrixConstructCheckedINTEL,
                   internal::OpCooperativeMatrixConstructCheckedINTEL)
_SPIRV_OP_INTERNAL(CooperativeMatrixLoadOffsetINTEL,
                   internal::OpCooperativeMatrixLoadOffsetINTEL)
_SPIRV_OP_INTERNAL(CooperativeMatrixStoreOffsetINTEL,
                   internal::OpCooperativeMatrixStoreOffsetINTEL)
_SPIRV_OP_INTERNAL(CooperativeMatrixApplyFunctionINTEL,
                   internal::OpCooperativeMatrixApplyFunctionINTEL)
_SPIRV_OP_INTERNAL(MaskedGatherINTEL, internal::OpMaskedGatherINTEL)
_SPIRV_OP_INTERNAL(MaskedScatterINTEL, internal::OpMaskedScatterINTEL)
_SPIRV_OP_INTERNAL(TaskSequenceAsyncINTEL, internal::OpTaskSequenceAsyncINTEL)
_SPIRV_OP_INTERNAL(TaskSequenceCreateINTEL, internal::OpTaskSequenceCreateINTEL)
_SPIRV_OP_INTERNAL(TaskSequenceGetINTEL, internal::OpTaskSequenceGetINTEL)
_SPIRV_OP_INTERNAL(TaskSequenceReleaseINTEL,
                   internal::OpTaskSequenceReleaseINTEL)
_SPIRV_OP_INTERNAL(TypeTaskSequenceINTEL, internal::OpTypeTaskSequenceINTEL)
_SPIRV_OP_INTERNAL(ConvertHandleToImageINTEL,
                   internal::ConvertHandleToImageINTEL)
_SPIRV_OP_INTERNAL(ConvertHandleToSamplerINTEL,
                   internal::ConvertHandleToSamplerINTEL)
_SPIRV_OP_INTERNAL(ConvertHandleToSampledImageINTEL,
                   internal::ConvertHandleToSampledImageINTEL)
_SPIRV_OP_INTERNAL(FSigmoidINTEL, internal::FSigmoidINTEL)
_SPIRV_OP_INTERNAL(ClampConvertFToFINTEL, internal::OpClampConvertFToFINTEL)
_SPIRV_OP_INTERNAL(StochasticRoundFToFINTEL,
                   internal::OpStochasticRoundFToFINTEL)
_SPIRV_OP_INTERNAL(ClampStochasticRoundFToFINTEL,
                   internal::OpClampStochasticRoundFToFINTEL)
_SPIRV_OP_INTERNAL(ClampConvertFToSINTEL,
                   internal::OpClampConvertFToSINTEL)
_SPIRV_OP_INTERNAL(ClampStochasticRoundFToSINTEL,
                   internal::OpClampStochasticRoundFToSINTEL)
_SPIRV_OP_INTERNAL(SubgroupScaledMatrixMultiplyAccumulateINTEL,
                   internal::OpSubgroupScaledMatrixMultiplyAccumulateINTEL)
